from pathlib import Path
from typing import Tuple, Type, Dict, Union, Optional

import numpy as np

import torch

from kekas.callbacks import OneCycleLR, Callbacks
# Callback, Callbacks, ProgressBarCallback, \
#     PredictionsSaverCallback, OneCycleLR, SimpleLossCallback, MetricsCallback, \
#     TBLogger, LRFinder, CheckpointSaverCallback, SimpleSchedulerCallback, \
#     EarlyStoppingCallback, SimpleOptimizerCallback, LRUpdater
from kekas.utils import DotDict
#     get_pbar, exp_weight_average, extend_postfix, to_numpy


# Implement annealing_cos
# https://github.com/fastai/fastai/blob/master/fastai/callbacks/one_cycle.py#L32


def kek_one_cycle(self,
                  max_lr: float,
                  cycle_len: int,
                  momentum_range: Tuple[float, float] = (0.95, 0.85),
                  div_factor: float = 25,
                  increase_fraction: float = 0.3,
                  opt: Optional[Type[torch.optim.Optimizer]] = None,
                  opt_params: Optional[Dict] = None,
                  logdir: Optional[Union[str, Path]] = None,
                  cp_saver_params: Optional[Dict] = None,
                  early_stop_params: Optional[Dict] = None,
                  annealing_cos: bool = False,
                  simulate: bool = False) -> None:
    """Kek your model to the moon with One Cycle policy!

    Conducts a one-cycle train-val procedure with several options for
    customization.
    For info about One Cycle policy please see:
    https://arxiv.org/abs/1803.09820
    https://sgugger.github.io/the-1cycle-policy.html

    Args:
        max_lr: the maximum learning rate that will be achieved during
            training process
        cycle_len: the number of full passes through the training dataset.
            It is quite similar to epochs number
        momentum_range: the range of optimizers momentum changes
        div_factor: is equal to max_lr / min_lr during one-cycle training
        increase_fraction: the fraction of the whole iterations during which
            the learning rate will increase
        opt: torch optimizer. If specified, than specified optimizer will be
            used for this train-val procedure, else the default one.
        opt_params: The kwargs dict for custom optimizer initialization.
            It should contain any opt params you want EXCEPT learning rate,
            Can be defined even for default optimizer.
        logdir: If provided, the TBLogger will be created and tensorboard
            logs will be written in this directory.
            For more info see kekas.callbacks.TBLogger and example ipynb's.
        cp_saver_params: kwargs dict parameters for CheckpointSaverCallback.
            If provided, then a CheckpointSaverCallback will be created.
            For more info see kekas.callbacks.CheckpointSaverCallback
            and example ipynb's.
        early_stop_params: kwargs dict parameters for EarlyStoppingCallback.
            If provided, then a EarlyStoppingCallback will be created.
            For more info see kekas.callbacks.EarlyStoppingCallback
            and example ipynb's.
    """

    callbacks = self.callbacks

    # temporarily add OneCycle callback
    len_loader = len(self.state.core.dataowner.train_dl)
    if annealing_cos:
        OCLR = OneCycleLR_AnnealingCos
    else:
        OCLR = OneCycleLR
    one_cycle_cb = OCLR(max_lr, cycle_len, len_loader,
                        momentum_range, div_factor, increase_fraction)

#   TODO: update on train begin
#    if one_cycle_cb.cut_point == 0:
#        one_cycle_cb.cut_point = 1
#        assert one_cycle_cb.total_iter - one_cycle_cb.cut_point != 0

    if simulate:
        from matplotlib import pyplot as plt
        one_cycle_cb.on_train_begin(None)
        xx = np.arange(one_cycle_cb.total_iter)
        yy = []
        yy_m = []
        for x in xx:
            one_cycle_cb.cycle_iter = x
            yy.append(one_cycle_cb.calc_lr())
            yy_m.append(one_cycle_cb.calc_momentum())
        fig, axes = plt.subplots(ncols=2, figsize=(16, 8))
        ax = axes[0]
        ax.plot(xx, yy)
        ax.set_xlabel('batch')
        ax.set_ylabel('LR')
        ax.set_title('LR')
        ax = axes[1]
        ax.plot(xx, yy_m)
        ax.set_xlabel('batch')
        ax.set_ylabel('momentum')
        ax.set_title('momentum')
        return {'batch_i': xx, 'LR': yy, 'mometum': yy_m}

    else:
        try:
            self.callbacks = Callbacks(callbacks.callbacks + [one_cycle_cb])

            self.kek(lr=max_lr,
                     epochs=cycle_len,
                     opt=opt,
                     opt_params=opt_params,
                     logdir=logdir,
                     cp_saver_params=cp_saver_params,
                     early_stop_params=early_stop_params)
        finally:
            # set old callbacks without OneCycle
            self.callbacks = callbacks


class OneCycleLR_AnnealingCos(OneCycleLR):
    """

    BAsed on OneCycleLR

    https://github.com/belskikh/kekas/blob/master/kekas/callbacks.py
    https://github.com/fastai/fastai/blob/master/fastai/callbacks/one_cycle.py

    An learning rate updater
        that implements the CircularLearningRate (CLR) scheme.
    Learning rate is increased then decreased linearly.
    Inspired by
    https://github.com/fastai/fastai/blob/master/fastai/callbacks/one_cycle.py
    """

    def on_train_begin(self, state: DotDict) -> None:
        self.total_iter = self.len_loader * self.cycle_len - 1
        self.cut_point = int(self.total_iter * self.increase_fraction)

        if self.cut_point == 0:
            self.cut_point = 1
            assert self.total_iter - self.cut_point != 0

    def calc_lr(self) -> float:

        # calculate percent for learning rate change
        if self.cycle_iter <= self.cut_point:
            percent = self.cycle_iter / self.cut_point
        else:
            percent = 1 - (self.cycle_iter - self.cut_point) / (self.total_iter - self.cut_point)

        percent = self.annealing_cos(0, 1, percent)

        res = self.init_lr * (1 + percent * (self.div_factor - 1)) / self.div_factor
        return res

    def calc_momentum(self) -> float:
        if self.cycle_iter <= self.cut_point:
            percent = 1 - self.cycle_iter / self.cut_point

        else:
            percent = (self.cycle_iter - self.cut_point) / (self.total_iter - self.cut_point)

        percent = self.annealing_cos(0, 1, percent)

        res = self.momentum_range[1] + percent * (self.momentum_range[0] - self.momentum_range[1])
        return res

    def annealing_cos(self, start: int, end: int, pct: float) -> float:
        "Cosine anneal from `start` to `end` as pct goes from 0.0 to 1.0."
        cos_out = np.cos(np.pi * pct) + 1
        return end + (start - end) / 2 * cos_out



from typing import Any, Dict, Hashable, List, Optional, Union
import numpy as np
import warnings
from matplotlib import pyplot as plt
from IPython.display import display, FileLink, HTML
from pathlib import Path
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator, ScalarEvent

import plotly.graph_objs as go


def plot_lr_loss(logdir, fn_save=None):
    fig, ax = plt.subplots(figsize=(8, 8))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        train_scalars = get_tensorboard_scalars(logdir + '/train', metrics=['loss', 'lr'], step='batch')
        tm_loss = get_scatter(train_scalars, 'batch/loss', prefix='train/')
        tm_lr = get_scatter(train_scalars, 'batch/lr', prefix='train/')

    ax.plot(np.log10(tm_lr['y']), np.log10(tm_loss['y']))
    ax.set_title('LR finder results:')
    ax.set_xlabel('lg(lr)')
    ax.set_ylabel('lg(loss)')

    if fn_save is None:
        plt.show()
    else:
        plt.savefig(fn_save)
        plt.close(fig)
        display(FileLink(Path(fn_save)))
        # return Image(filename=fn_animation)
        display(HTML(f"<img src='{fn_save}' />"))


def plot_metrics(logdir, step="batch",
                 metrics: Union[Any, List, Hashable] = None,
                 log10=False, alpha=0.8, multiruns=True,
                 fn_save=None,
                 compare_with_logdir=None,
                 ):

    logdir = Path(logdir)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        train_scalars, val_scalars = get_train_val_scalars(logdir=logdir, metrics=metrics, step=step, multiruns=multiruns)

    if compare_with_logdir is not None:
        compare_with_logdir = Path(compare_with_logdir)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            train_scalars_2, val_scalars_2 = get_train_val_scalars(logdir=compare_with_logdir, metrics=metrics, step=step, multiruns=multiruns)

    n = len(train_scalars)
    fig, axes = plt.subplots(figsize=(8 * n, 8), ncols=n)
    if n == 1:
        axes = [axes]

    for i in range(n):
        m = list(train_scalars.keys())[i]
        ax = axes[i]

        tm = get_scatter(train_scalars, m, prefix="train/", multiruns=multiruns)
        try:
            vm = get_scatter(val_scalars, m, prefix="val/", multiruns=multiruns)
            data = [tm, vm]
        except:
            data = [tm]

        ax.set_title(m)
        for tm in data:
            x = tm['x']
            y = tm['y']
            if log10:
                y = np.log10(y)

            label = tm['name'].split('/')[0]

            ax.plot(x, y, label=label, alpha=alpha)
            # ax.plot(y, label=label, alpha=alpha)

        if compare_with_logdir is not None:
            m2 = list(train_scalars_2.keys())[i]
            tm = get_scatter(train_scalars_2, m2, prefix="train/", multiruns=multiruns)
            try:
                vm = get_scatter(val_scalars_2, m2, prefix="val/", multiruns=multiruns)
                data = [tm, vm]
            except:
                data = [tm]

            for tm in data:
                x = tm['x']
                y = tm['y']
                if log10:
                    y = np.log10(y)
                label = "other " + tm['name'].split('/')[0]
                ax.plot(x, y, label=label, alpha=alpha, linestyle="dotted")

        ax.set_xlabel(step)
        if log10:
            ax.set_ylabel(f'log10({m})')

        ax.legend()

    if fn_save is None:
        plt.show()
    else:
        plt.savefig(fn_save)
        plt.close(fig)
        display(FileLink(Path(fn_save)))
        # return Image(filename=fn_animation)
        src = f"{fn_save}?{np.random.random()}"
        display(HTML(f"<img src='{src}' />"))


def get_train_val_scalars(logdir, metrics, step, multiruns):
    if multiruns:
        train_scalars = get_tensorboard_scalars_multiruns(logdir / 'train', metrics=metrics, step=step)
        val_scalars = get_tensorboard_scalars_multiruns(logdir / 'val', metrics=metrics, step=step)
    else:
        train_scalars = get_tensorboard_scalars(logdir / 'train', metrics=metrics, step=step)
        val_scalars = get_tensorboard_scalars(logdir / 'val', metrics=metrics, step=step)
    return train_scalars, val_scalars


# from kekas.utils
def get_tensorboard_scalars(
    logdir: str, metrics: Optional[List[str]], step: str
) -> Dict[str, List]:
    event_acc = EventAccumulator(str(logdir))
    event_acc.Reload()

    if metrics is not None:
        scalar_names = [
            n for n in event_acc.Tags()["scalars"] if step in n and any(m in n for m in metrics)
        ]
    else:
        scalar_names = [n for n in event_acc.Tags()["scalars"] if step in n]

    scalars = {sn: event_acc.Scalars(sn) for sn in scalar_names}
    return scalars


def get_tensorboard_scalars_multiruns(
    logdir: str, metrics: Optional[List[str]], step: str
) -> Dict[str, List]:

    scalars = get_tensorboard_scalars(logdir, metrics, step)
    # print(scalars)
    return scalars


def get_scatter(scalars: Dict[str, ScalarEvent], name: str, prefix: str, multiruns=False) -> go.Scatter:
    xs = [s.step for s in scalars[name]]
    if multiruns:
        xs = list(range(0, len(xs)))
    ys = [s.value for s in scalars[name]]

    return go.Scatter(x=xs, y=ys, name=prefix + name)



import numpy as np
import pandas as pd
import humanize
from collections import OrderedDict
from IPython.display import display
import torch
import torch.nn as nn


def print_trainable_parameters2(model, max_rows=400, max_colwidth=100, with_class=False):

    d = get_trainable_parameters(model, with_class=with_class)
    print('Total:', humanize.intcomma(d['total_number']))
    with pd.option_context("display.max_rows", max_rows, 'max_colwidth', max_colwidth):
        display(d['df_params'])


def get_trainable_parameters(model, with_class=False):
    params = []
    total_number = 0
    if with_class:
        gen = named_parameters2(model)
    else:
        gen = model.named_parameters()
    for name, p, module in gen:
        if with_class:
            name, class_name = name

        requires_grad = p.requires_grad

        shape = " x ".join([str(i) for i in list(p.shape)])
        number = p.numel()
        p_np = p.cpu().detach().numpy()
        d = {'name': name, 'shape': shape, 'number': number, 'sum':p_np.sum(), 'mean':p_np.mean(), 'requires_grad':requires_grad}
        if with_class:
            d['class'] = class_name
        params.append(d)
        total_number += number
    df_params = pd.DataFrame(params)
    return {'df_params': df_params, 'total_number': total_number}


def named_parameters2(self, prefix='', recurse=True):
    r"""Returns an iterator over module parameters, yielding both the
    name of the parameter as well as the parameter itself.

    Args:
        prefix (str): prefix to prepend to all parameter names.
        recurse (bool): if True, then yields parameters of this module
            and all submodules. Otherwise, yields only parameters that
            are direct members of this module.

    Yields:
        (string, Parameter): Tuple containing the name and parameter

    Example::

        >>> for name, param in self.named_parameters():
        >>>    if name in ['bias']:
        >>>        print(param.size())

    """
    gen = _named_members2(
        self,
        lambda module: module._parameters.items(),
        prefix=prefix, recurse=recurse)
    for elem in gen:
        yield elem


def _named_members2(self, get_members_fn, prefix='', recurse=True):
    r"""Helper method for yielding various names + members of modules."""
    memo = set()
    modules = self.named_modules(prefix=prefix) if recurse else [(prefix, self)]
    for module_prefix, module in modules:
        members = get_members_fn(module)
        for k, v in members:
            if v is None or v in memo:
                continue
            memo.add(v)
            name = module_prefix + ('.' if module_prefix else '') + k
            class_name = fullname(module)
            class_name = module.__class__.__name__
            yield (name, class_name), v, module


def fullname(o):
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.

    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__  # Avoid reporting __builtin__
    else:
        return module + '.' + o.__class__.__name__


def attach_names_to_modules(model):
    for a in named_parameters2(model):
        pname = a[0][0]
        # class_name = a[0][1]
        # param  = a[1]
        module = a[2]

        module_name = '.'.join(pname.split('.')[:-1])
        module.__name = module_name


def get_object_size(o):
    res = [None]
    if hasattr(o, 'get_spatial_locations'):
        res = [list(o.features.size()),
               list(o.get_spatial_locations().size()),
               list(o.spatial_size.numpy())]
    elif hasattr(o, 'size'):
        res = list(o.size())
    return res


def summary(model, xb=None, input_size=None, batch_size=-1, device="cuda", verbose=0, return_string=False, return_data=False):
    # https://github.com/sksq96/pytorch-summary/blob/master/torchsummary/torchsummary.py
    if xb is None:
        assert input_size is not None
    if input_size is None:
        assert xb is not None

    attach_names_to_modules(model)

    def register_hook(module):

        def hook(module, input, output):
            class_name = str(module.__class__).split(".")[-1].split("'")[0]
            full_class_name = fullname(module)
            module_idx = len(summary)

            if verbose:
                print(type(module))
                print('module:', full_class_name)
                print('output:', fullname(output))

            m_key = "%s-%i" % (class_name, module_idx + 1)
            summary[m_key] = OrderedDict()

            summary[m_key]['class_name'] = class_name
            summary[m_key]['idx'] = module_idx + 1
            summary[m_key]['name'] = getattr(module, '__name', '')

            if full_class_name == 'sparseconvnet.ioLayers.InputLayer':
                # coords and features
                cf = input[0]
                summary[m_key]["input_shape"] = list(cf[0].size()), list(cf[1].size())
            elif isinstance(input[0], dict):
                cf = input[0]
                summary[m_key]["input_shape"] = [list(v.size()) for key, v in cf.items() if isinstance(v, torch.Tensor)]
            # elif full_class_name.startswith('sparseconvnet.') or  full_class_name.startswith('SCN_architectures'):
            elif hasattr(input[0], 'get_spatial_locations'):
                summary[m_key]["input_shape"] = (list(input[0].features.size()),
                                                 list(input[0].get_spatial_locations().size()),
                                                 list(input[0].spatial_size.numpy()))
            elif isinstance(input[0], (list, tuple)):
                summary[m_key]["input_shape"] = [get_object_size(o) for o in input[0]]
            else:
                if verbose > 1:
                    print(full_class_name, input[0])
                summary[m_key]["input_shape"] = list(input[0].size())
                # summary[m_key]["input_shape"][0] = batch_size

            if isinstance(output, (list, tuple)):
                summary[m_key]["output_shape"] = [
                    [-1] + get_object_size(o)[1:] for o in output
                ]
            else:
                # if fullname(output) in ['sparseconvnet.sparseConvNetTensor.SparseConvNetTensor', 'sparseconvnet.ioLayers.InputLayer']:
                if hasattr(output, 'get_spatial_locations'):
                    summary[m_key]["output_shape"] = (list(output.get_spatial_locations().size()),
                                                      list(output.features.size()),
                                                      list(output.spatial_size.numpy()))
                else:
                    summary[m_key]["output_shape"] = list(output.size())
                    # summary[m_key]["output_shape"][0] = batch_size

            params = 0
            if hasattr(module, "weight") and hasattr(module.weight, "size"):
                weight_size = list(module.weight.size())
                summary[m_key]["weight_size"] = weight_size
                params += torch.prod(torch.LongTensor(weight_size))
                summary[m_key]["trainable"] = module.weight.requires_grad
            if hasattr(module, "bias") and hasattr(module.bias, "size"):
                bias_size = list(module.bias.size())
                summary[m_key]["bias_size"] = bias_size
                params += torch.prod(torch.LongTensor(bias_size))
            summary[m_key]["nb_params"] = params
        # print("module", module)
        # print(isinstance(module, nn.Sequential))
        # print(module == model)
        if not isinstance(module, nn.Sequential) \
                and not isinstance(module, nn.ModuleList):
            # and not (module == model)
            hooks.append(module.register_forward_hook(hook))

    device = device.lower()
    assert device in [
        "cuda",
        "cpu",
    ], "Input device is not valid, please specify 'cuda' or 'cpu'"

    if device == "cuda" and torch.cuda.is_available():
        dtype = torch.cuda.FloatTensor
    else:
        dtype = torch.FloatTensor

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # batch_size of 2 for batchnorm
    if xb is None:
        x = [torch.rand(2, *in_size).type(dtype) for in_size in input_size]
    # print(type(x[0]))

    # create properties
    summary = OrderedDict()
    hooks = []

    # register hook
    model.apply(register_hook)

    try:
        # make a forward pass
        # print(x.shape)
        if xb is None:
            model(*x)
        elif isinstance(xb, dict):
            model(xb)
        else:
            model(*xb)

    finally:
        # remove these hooks
        for h in hooks:
            h.remove()

    lines = []

    def add_line(line):
        lines.append(line)

    add_line("-" * 90)
    line_new = "{:>30}  {:>35} {:>15}".format("Layer (type)", "Output Shape", "Param #")
    add_line(line_new)
    line_new = "{:>30}  {:>35} {:>15}".format("", "coords   features  spatial_size", "")
    add_line(line_new)
    add_line("=" * 90)
    total_params = 0
    total_output = 0
    trainable_params = 0
    # print(summary.keys())
    for layer in summary:
        # input_shape, output_shape, trainable, nb_params
        line_new = "{:>30}  {:<30} {:>35} {:>15}".format(
            layer,
            str(summary[layer]["name"]),
            str(summary[layer]["output_shape"]),
            "{0:,}".format(summary[layer]["nb_params"]),
        )
        total_params += summary[layer]["nb_params"]
        try:
            # TODO:
            total_output += np.prod(summary[layer]["output_shape"])
        except:
            pass
        if "trainable" in summary[layer]:
            if summary[layer]["trainable"] is True:
                trainable_params += summary[layer]["nb_params"]
        add_line(line_new)

    # assume 4 bytes/number (float on cuda).
    # total_input_size = abs(np.prod(input_size) * batch_size * 4. / (1024 ** 2.))
    # total_output_size = abs(2. * total_output * 4. / (1024 ** 2.))  # x2 for gradients
    # total_params_size = abs(total_params.numpy() * 4. / (1024 ** 2.))
    # total_size = total_params_size + total_output_size + total_input_size

    add_line("=" * 90)
    add_line("Total params: {0:,}".format(total_params))
    add_line("Trainable params: {0:,}".format(trainable_params))
    add_line("Non-trainable params: {0:,}".format(total_params - trainable_params))
    add_line("-" * 90)
    # print("Input size (MB): %0.2f" % total_input_size)
    # print("Forward/backward pass size (MB): %0.2f" % total_output_size)
    # print("Params size (MB): %0.2f" % total_params_size)
    # print("Estimated Total Size (MB): %0.2f" % total_size)
    # print("----------------------------------------------------------------")
    # return summary
    s = '\n'.join(lines)
    if return_string:
        return s
    if return_data:
        return list(summary.values())
    else:
        print(s)



"A `Callback` that saves tracked metrics into a Neptune"
# Contribution from devforfu: https://nbviewer.jupyter.org/gist/devforfu/ea0b3fcfe194dad323c3762492b05cae

from collections import defaultdict
import numpy as np
import time

from kekas.callbacks import Callback
from kekas.utils import DotDict, get_opt_lr

__all__ = ['NeptuneMonitor']


class NeptuneMonitor(Callback):
    "A `Callback` that send history of metrics to `neptune`."

    def __init__(self, neptune_experiment, simulation=False):
        self.neptune_experiment = neptune_experiment
        self.total_iter = 0
        self.train_iter = 0
        self.val_iter = 0
        self.train_batch_iter = 0
        self.val_batch_iter = 0
        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)

        self.epochs_total = 1
        self.simulation = simulation

    def update_total_iter(self, mode: str) -> None:
        if mode == "train":
            self.train_iter += 1
            self.train_batch_iter += 1
        if mode == "val":
            self.val_iter += 1
            self.val_batch_iter += 1
            self.total_iter += 1    # ?????

    def on_train_begin(self, state: DotDict) -> None:
        self.train_iter = 0
        self.val_iter = 0

        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict):
        self.train_batch_iter = 0
        self.val_batch_iter = 0
        self.val_t0 = time.time()
        self.train_t0 = time.time()

    def on_batch_end(self, i: int, state: DotDict) -> None:
        if state.core.mode == "train":
            for name, metric in state.core.metrics["train"].items():
                self.train_metrics[name].append(float(metric))

#                lr = get_opt_lr(state.core.opt)
#                self.train_writer.add_scalar("batch/lr",
#                                             float(lr),
#                                             global_step=self.train_iter)

            self.update_total_iter(state.core.mode)

        elif state.core.mode == "val":
            for name, metric in state.core.metrics["val"].items():
                self.val_metrics[name].append(float(metric))

            self.update_total_iter(state.core.mode)

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:

        if state.core.mode == "train":
            for name, metric in self.train_metrics.items():
                mean = np.mean(metric[-self.train_batch_iter:])
                self.send_metric(name, self.epochs_total, mean)

            self.send_metric('epoch_time_train', self.epochs_total, time.time() - self.train_t0)

        if state.core.mode == "val":

            for name, metric in self.val_metrics.items():
                mean = np.mean(metric[-self.val_batch_iter:])  # last epochs vals
                self.send_metric("val_" + name, self.epochs_total, mean)

            lr = get_opt_lr(state.core.opt)
            self.send_metric("LR", self.epochs_total, lr)

            self.epochs_total += 1

            self.send_metric('epoch_time_valid', self.epochs_total, time.time() - self.val_t0)

    def send_metric(self, name, x, y):
        if not self.simulation and (self.neptune_experiment is not None):
            self.neptune_experiment.send_metric(name, x, y)
        else:
            print('\n', name, x, y)



import inspect
import warnings
from pathlib import Path


def send_model_code(model, model_config, logdir, NEPTUNE_ON=False, exp=None):
    model_init = None
    model_forward = None
    model_config_s = None
    try:
        model_init = inspect.getsource(model.__init__)
    except Exception as e:
        warnings.warn(f"Can't save model_init: {e}", UserWarning)
    try:
        model_forward = inspect.getsource(model.forward)
    except Exception as e:
        warnings.warn(f"Can't save model_forward: {e}", UserWarning)

    try:
        model_config_s = str(model_config)
    except Exception as e:
        warnings.warn(f"Can't save model_config: {e}", UserWarning)

    def save_and_send(src, fnbase):
        if src is not None:
            fn = Path(logdir) / fnbase
            with open(fn, 'w') as f:
                f.write(src)
            if NEPTUNE_ON and exp is not None:
                exp.send_artifact(fn)

    save_and_send(model_init, 'model_init.py')
    save_and_send(model_forward, 'model_forward.py')
    save_and_send(model_config_s, 'model_config.txt')


def log_and_send_string(value, name='example.txt', logdir=None, NEPTUNE_ON=False, exp=None):

    def save_and_send(src, fnbase):
        if src is not None:
            fn = Path(logdir) / fnbase
            with open(fn, 'w') as f:
                f.write(src)
            if NEPTUNE_ON and exp is not None:
                exp.send_artifact(fn)

    save_and_send(value, name)



from collections import defaultdict
from pathlib import Path
from typing import Union

import numpy as np
from tensorboardX import SummaryWriter
from kekas.utils import DotDict, get_opt_lr

from kekas.callbacks import Callback

__all__ = ['TBLogger']


class TBLogger(Callback):
    def __init__(self, logdir: Union[str, Path]) -> None:
        self.logdir = Path(logdir)
        self.writer = None
        self.total_iter = 0
        self.train_iter = 0
        self.val_iter = 0
        self.train_batch_iter = 0
        self.val_batch_iter = 0
        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)
        self.total_epoch_iter = 0

    def update_total_iter(self, mode: str) -> None:
        if mode == "train":
            self.train_iter += 1
            self.train_batch_iter += 1
        if mode == "val":
            self.val_iter += 1
            self.val_batch_iter += 1
        self.total_iter += 1

    def on_train_begin(self, state: DotDict) -> None:
        # self.train_iter = 0
        # self.val_iter = 0
        self.logdir.mkdir(exist_ok=True)
        self.train_writer = SummaryWriter(str(self.logdir / "train"))
        self.val_writer = SummaryWriter(str(self.logdir / "val"))
        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)

    def on_epoch_begin(self, epoch: int, epochs: int, state: DotDict):
        self.train_batch_iter = 0
        self.val_batch_iter = 0

    def on_batch_end(self, i: int, state: DotDict) -> None:
        if state.core.mode == "train":
            for name, metric in state.core.metrics["train"].items():
                self.train_writer.add_scalar(f"batch/{name}",
                                             float(metric),
                                             global_step=self.total_iter)
                self.train_metrics[name].append(float(metric))

            lr = get_opt_lr(state.core.opt)
            self.train_writer.add_scalar("batch/lr",
                                         float(lr),
                                         global_step=self.train_iter)

            self.update_total_iter(state.core.mode)

        elif state.core.mode == "val":
            for name, metric in state.core.metrics["val"].items():
                self.val_writer.add_scalar(f"batch/{name}",
                                           float(metric),
                                           global_step=self.total_iter)
                self.val_metrics[name].append(float(metric))

            self.update_total_iter(state.core.mode)

    def on_epoch_end(self, epoch: int, state: DotDict) -> None:

        epoch = self.total_epoch_iter

        if state.core.mode == "train":
            for name, metric in self.train_metrics.items():
                mean = np.mean(metric[-self.train_batch_iter:])
                self.train_writer.add_scalar(f"epoch/{name}",
                                             float(mean),
                                             global_step=epoch)
        if state.core.mode == "val":
            for name, metric in self.val_metrics.items():
                mean = np.mean(metric[-self.val_batch_iter:])  # last epochs vals
                self.val_writer.add_scalar(f"epoch/{name}",
                                           float(mean),
                                           global_step=epoch)

            for name, param in state.core.model.named_parameters():
                self.val_writer.add_histogram(name, param.clone().cpu().data.numpy(), epoch)

            self.total_epoch_iter += 1

    def on_train_end(self, state: DotDict) -> None:
        self.train_writer.close()
        self.val_writer.close()



import os
import sys
import subprocess
import collections
from collections import OrderedDict
import platform
import re
import git
import warnings
import tempfile
import numpy as np
import functools


def dict_flatten(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(dict_flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def wide_notebook(percents=70):
    from IPython.display import display, HTML
    if percents < 30:
        percents = 30
    display(HTML("<style>.container { width:" + str(percents) + "% !important; }</style>"))


def watermark(packages=['python', 'virtualenv', 'nvidia', 'cudnn', 'hostname', 'torch'], return_string=True):
    lines = OrderedDict()
    if 'virtualenv' in packages:
        r = None
        if 'PS1' in os.environ:
            r = os.environ['PS1']
        elif 'VIRTUAL_ENV' in os.environ:
            r = os.environ['VIRTUAL_ENV']
        lines['virtualenv'] = r
    if 'python' in packages:
        r = sys.version.splitlines()[0]
        m = re.compile(r'([\d\.]+)').match(r)
        if m:
            r = m.groups()[0]
        lines['python'] = r
    if 'hostname' in packages:
        lines["hostname"] = platform.node()

    def find_in_lines(pip_list, package_name, remove_name=True):
        res = ''
        for line in pip_list:
            if hasattr(line, 'decode'):
                line = line.decode('utf-8')
            if package_name in line and line.startswith(package_name):
                if remove_name:
                    res = package_name.join(line.split(package_name)[1:]).strip()
                else:
                    res = line.strip()
                break
        return res

    if 'nvidia' in packages:
        lines['nvidia driver'] = subprocess.check_output(
            ["nvidia-smi", "--query-gpu=driver_version", "--format=csv,noheader"]).splitlines()[0]
        try:
            r = subprocess.check_output(["nvcc", "--version"]).splitlines()
            r = find_in_lines(r, 'release', False)
            r = r.split('release')[1].strip()
            lines['nvidia cuda'] = r
        except:
            pass

    if ('cudnn' in packages) and sys.platform.startswith('linux'):
        try:
            with open('/usr/local/cuda/include/cudnn.h', 'r') as f:
                r = f.readlines()
            v1 = find_in_lines(r, '#define CUDNN_MAJOR')
            v2 = find_in_lines(r, '#define CUDNN_MINOR')
            v3 = find_in_lines(r, '#define CUDNN_PATCHLEVEL')
            lines['cudnn'] = "{}.{}.{}".format(v1, v2, v3)
        except:
            pass

    pip_list = subprocess.check_output(["pip", "list"]).splitlines()

    if 'keras' in packages:
        lines['keras'] = find_in_lines(pip_list, 'Keras')

    if 'tensorflow' in packages:
        lines['tensorflow-gpu'] = find_in_lines(pip_list, 'tensorflow-gpu')

    if 'torch' in packages:
        lines['torch'] = find_in_lines(pip_list, 'torch')

    # with git commit hash
    for key in ['sparseconvnet', 'pytorch-lightning']:
        if key in packages:
            line = find_in_lines(pip_list, key)
            try:
                a = re.split(r'\s+', line)
                if len(a) > 1:
                    path = ' '.join(a[1:])
                    repo = git.Repo(path)
                    h = repo.head.object.hexsha
                    line = f"{line} {h}"
            except:
                pass
            lines[key] = line

    parsed = ['python', 'virtualenv', 'nvidia', 'cudnn', 'hostname', 'sparseconvnet', 'pytorch-lightning']
    for key in packages:
        if key not in parsed:
            lines[key] = find_in_lines(pip_list, key)

    res = ["{: <15} {}".format(k + ":", v) for (k, v) in lines.items()]

    s = "\n".join(res)
    print(s)

    if return_string:
        return s


def exec_and_print(command):
    c = command.split(' ')
    MyOut = subprocess.Popen(
        c,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True)
    stdout, stderr = MyOut.communicate()
    print(stdout)
    print(stderr)


def get_gpu_memory_map():
    """Get the current gpu usage.

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """
    result = subprocess.check_output(
        [
            'nvidia-smi', '--query-gpu=memory.used',
            '--format=csv,nounits,noheader'
        ])
    # print(result)
    result = result.decode('utf8')
    # Convert lines into a dictionary
    gpu_memory = [int(x) for x in result.strip().split('\n')]
    gpu_memory_map = dict([("gpu_{}".format(i), v) for i, v in enumerate(gpu_memory)])
    # gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map


def get_gpu_names_map():
    """Get the current gpus name.

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """
    result = subprocess.check_output(
        [
            'nvidia-smi', '--query-gpu=name',
            '--format=csv,nounits,noheader'
        ])
    # print(result)
    result = result.decode('utf8')
    # Convert lines into a dictionary
    gpu_names = [x for x in result.strip().split('\n')]
    gpu_names = dict([("gpu_{}".format(i), v) for i, v in enumerate(gpu_names)])
    # gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_names


def gpu_name_by_n(n):
    gpu_names = get_gpu_names_map()
    return gpu_names.get('gpu_{}'.format(n), None)


# TODO: move to trackers
def log_text_as_artifact(tracker, text, destination=None, existed_temp_file=None):
    fd = None
    try:
        fd, path = tempfile.mkstemp()

        with os.fdopen(fd, 'w') as tmp:
            tmp.write(text)

        try:
            tracker.log_artifact(path, destination)
        except Exception as e:
            warnings.warn(f"Can't .log_artifact for tracker {tracker}. {e}", UserWarning)
    except Exception as e:
        warnings.warn(f"Can't .log_text_as_artifact for composer {e}", UserWarning)
    finally:
        if fd is not None:
            os.remove(path)


def is_notebook():
    try:
        # pylint: disable=pointless-statement,undefined-variable
        get_ipython         # noqa: F821
        return True
    except Exception:
        return False


def rgetattr(obj, path: str, *default):
    """
    Get object attribute recursively
    :param obj: Object
    :param path: 'attr1.attr2.etc'
    :param default: Optional default value, at any point in the path
    :return: obj.attr1.attr2.etc

    https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties
    """
    DELIMITER = "."
    attrs = path.split(DELIMITER)
    try:
        return functools.reduce(getattr, attrs, obj)
    except AttributeError:
        if default:
            return default[0]
        raise

def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


# TODO: move to sparse


def count_voxels(coords):
    # TODO: use numpy.unique
    df = pd.DataFrame(coords, columns=['x', 'y', 'z'])
    n_voxels = len(df.groupby(['x', 'y', 'z']).count())
    return n_voxels


def split_coords_features(coords_features):
    coords = coords_features[0]
    features = coords_features[1]

    assert coords.shape[1] == 4
    example_indices = coords[:, 3]
    N = example_indices.max() + 1

    res_coors_features = []

    for i in range(N):
        indices = (example_indices == i)
        coords_i = coords[indices, :3]
        features_i = features[indices]
        res_coors_features.append([coords_i, features_i])
    return res_coors_features


def sparse_to_dense_sgnn(coords, sdf, dims, fillna=np.inf):
    """
    Warning sdf -->  - sdf
    """
    res = np.ones(dims, dtype=np.float32)
    res = res * fillna
    res[coords[:, 0], coords[:, 1], coords[:, 2]] = -sdf
    res = np.expand_dims(res, axis=0)
    return res



# from pathlib import Path
# import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict
from typing import List
from goodok_mlu.utils import rgetattr

params_aliases = [
    {"old_key_starts": 'surfacepred.p1.',
     "new_key_starts": "surfacepred.fcn.block.1."},
    {"old_key_starts": 'surfacepred.p2.',
     "new_key_starts": "surfacepred.fcn.block.2."},
    {"old_key_starts": 'surfacepred.p3.',
     "new_key_starts": "surfacepred.fcn.block.3."},

    {"old_key_starts": 'refinement.0.p1.',
     "new_key_starts": "refinement.0.fcn.block.1."},
    {"old_key_starts": 'refinement.0.p1.',
     "new_key_starts": "refinement.0.fcn.block.1."},
    {"old_key_starts": 'refinement.0.p2.',
     "new_key_starts": "refinement.0.fcn.block.2."},
    {"old_key_starts": 'refinement.0.p3.',
     "new_key_starts": "refinement.0.fcn.block.3."},
    {"old_key_starts": 'refinement.1.p1.',
     "new_key_starts": "refinement.1.fcn.block.1."},
    {"old_key_starts": 'refinement.1.p1.',
     "new_key_starts": "refinement.1.fcn.block.1."},
    {"old_key_starts": 'refinement.1.p2.',
     "new_key_starts": "refinement.1.fcn.block.2."},
    {"old_key_starts": 'refinement.1.p3.',
     "new_key_starts": "refinement.1.fcn.block.3."},
    {"old_key_starts": 'refinement.2.p1.',
     "new_key_starts": "refinement.2.fcn.block.1."},
    {"old_key_starts": 'refinement.2.p1.',
     "new_key_starts": "refinement.2.fcn.block.1."},
    {"old_key_starts": 'refinement.2.p2.',
     "new_key_starts": "refinement.2.fcn.block.2."},
    {"old_key_starts": 'refinement.2.p3.',
     "new_key_starts": "refinement.2.fcn.block.3."},
]


def model_load_state_dict(model, model_path=None, state=None):
    if model_path is not None:
        state = torch.load(model_path)
    state = replace_aliases(state, params_aliases=params_aliases)
    model.load_state_dict(state)
    return model


def replace_aliases(state, params_aliases=params_aliases):
    res = OrderedDict()
    for key in state.keys():
        a = get_alias(key, params_aliases)
        if a is not None:
            v = state[key]
            res[a['new_key']] = v
        else:
            res[key] = state[key]
    return res


def get_alias(key, params_aliases):
    res = None
    for a in params_aliases:
        if key.startswith(a['old_key_starts']):
            res = a.copy()
            res['new_key'] = key.replace(a['old_key_starts'], a['new_key_starts'])
            break
        elif key.startswith('model.' + a['old_key_starts']):
            res = a.copy()
            res['new_key'] = key.replace('model.' + a['old_key_starts'], 'model.' + a['new_key_starts'])
            break
    return res


def load_model_variants(pl_model, model, model_path):
    d = torch.load(model_path)
    if len(d.keys()) == 3:
        # original, author's weights, or running original authors code
        d = torch.load(model_path)
        print(d.keys())
        print()
        for k in ['epoch']:
            s = f'{k}:'
            print(f'{s:<30} {d.get(k, None)}')
        print()
        state = replace_aliases(d['state_dict'], params_aliases=params_aliases)
        model.load_state_dict(state)
        # del d
    elif model_path.suffix == '.ckpt':
        # pytorch-lightning
        d = torch.load(model_path)
        print(d.keys())
        for k in ['epoch', 'global_step', 'pytorch-lightning_version', 'checkpoint_callback_best_model_score', 'checkpoint_callback_best_model_path',
                  'hparams_name', 'hparams_type']:
            s = f'{k}:'
            print(f'{s:<30} {d.get(k, None)}')
        print()

        # log('state_dict files', d['state_dict']['model.encoder.process_sparse.0.p1.weight'])
        # pl_model.model.load_state_dict(d['state_dict'])

        state = replace_aliases(d['state_dict'], params_aliases=params_aliases)

        pl_model.load_state_dict(state)

        # pl_model = pl_model.load_from_checkpoint(str(model_path))
        # model = pl_model.model
        # log('state_dict after', pl_model.model.state_dict()['encoder.process_sparse.0.p1.weight'])
        # log('state_dict after2',model2.state_dict()['model.encoder.process_sparse.0.p1.weight'])

        del d
    else:
        # pure model dict_state
        state = torch.load(model_path)
        model.load_state_dict(state)

    return pl_model, model


def count_params(model):
    return sum(p.numel() for p in model.parameters())


# from kekas

def set_grad(module: nn.Module, requires_grad: bool) -> None:
    for param in module.parameters():
        param.requires_grad = requires_grad


def freeze_by_names(module: nn.Module, modules_to_freeze : List[str]) -> None:
    """
        freeze parts of module by list of submodule names, recursively
        Examples:
            freeze_by_names(model, ['unet.Decoder', 'unet.Encoder']) - freeze all nested layers of decoder and encoder
            freeze_by_names(model, ['model.encoder.process_sparse.0'])
    """
    prefix = ""
    for s in modules_to_freeze:
        print("\n", f'freeze module "{s}"', "\n")
        m_to_freeze = rgetattr(module, s)
        modules = m_to_freeze.named_modules(prefix=prefix)
        for i, named_m in enumerate(modules):
            name = named_m[0]
            m = named_m[1]
            set_grad(m, requires_grad=False)


def sparse_to_tensors(sparse):
    """SparseConvNetTensor --> (locs/coords, features, spatial_size)"""
    return sparse.get_spatial_locations(), sparse.features, sparse.spatial_size



from pathlib import Path
import numpy as np
import torch
import json


path_root = DIR_THIS = Path('.').absolute()


def full_path(fn, path=None):
    if path is None:
        return Path(fn)
    else:
        return Path(path) / fn


def load_tensor(fn, to_cuda=False, path=None):
    fn = full_path(fn, path)
    fn = fn.with_suffix('.npy')
    res = np.load(fn)
    res = torch.from_numpy(res)
    if to_cuda:
        res = res.cuda()
    return res


def load_tensor_list(fn, to_cuda=False, path=None, n=4):
    fn = full_path(fn, path)
    res = []
    for i in range(n):
        fn_i = fn.with_suffix(f'.{i}.npy')
        v = np.load(fn_i)
        v = torch.from_numpy(v)
        if to_cuda:
            v = v.cuda()
        res.append(v)
    return res


def load_coords_features(fn, features_to_device=False, path=None):
    fn = full_path(fn, path)
    fn1 = fn.with_suffix('.coords.npy')
    fn2 = fn.with_suffix('.features.npy')

    coords = torch.from_numpy(np.load(fn1))
    features = torch.from_numpy(np.load(fn2))
    if features_to_device:
        features = features.cuda()

    return [coords, features]


def save_tensor(fn, x, from_cuda=False, path=None, verbose=False):
    fn = full_path(fn, path)
    fn = fn.with_suffix('.npy')
    if from_cuda:
        x = x.cpu().detach().numpy()
    np.save(fn, x)
    if verbose:
        print(f'{fn} saved')


def save_tensor_list(fn, x, from_cuda=False, path=None, verbose=False):
    fn = full_path(fn, path)
    for i, v in enumerate(x):
        fn_i = fn.with_suffix(f'.{i}.npy')
        if from_cuda:
            v = v.cpu().detach().numpy()
        np.save(fn_i, v)

        if verbose:
            print(f'{fn_i} saved')


def save_coords_features(fn, coords_features, spatial_size=None, path=None, verbose=False):
    fn = full_path(fn, path)
    fn1 = fn.with_suffix('.coords.npy')
    fn2 = fn.with_suffix('.features.npy')

    np.save(fn1, coords_features[0].cpu().detach())
    np.save(fn2, coords_features[1].cpu().detach())

    if verbose:
        print(f'{fn1} saved')
        print(f'{fn2} saved')

    if spatial_size is not None:
        fn3 = fn.with_suffix('.size.npy')
        np.save(fn3, spatial_size)
        if verbose:
            print(f'{fn3} saved')


def save_json(fn, x, path=None, verbose=False):
    fn = full_path(fn, path)
    fn = fn.with_suffix('.json')
    with open(fn, "w") as f:
        json.dump(x, f)
    if verbose:
        print(f'{fn} saved')



import numpy as np
import warnings


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_color(text, color='green', bold=False, underline=False):
    assert color in ['green', 'blue', 'red', 'olive', 'magenta']
    if color == 'green':
        print(bcolors.OKGREEN, end='')
    elif color == 'blue':
        print(bcolors.OKBLUE, end='')
    elif color == 'red':
        print(bcolors.FAIL, end='')
    elif color == 'olive':
        print(bcolors.WARNING, end='')
    elif color == 'magenta':
        print(bcolors.HEADER, end='')

    if bold:
        print(bcolors.BOLD, end='')
    if underline:
        print(bcolors.UNDERLINE, end='')

    print(text, end='')

    print(bcolors.ENDC)


log_options = {
    'max_name_length': 14,
    'max_shape_length': 14,
}

scalar_types = (int, np.int, np.int32, np.int64, np.uint, np.uint32, np.uint64)


def log(text, array=None, indent=''):
    """Prints a text message. And, optionally, if a Numpy array is provided iterators
    prints iterators's shape, min, and max values.
    """
    if isinstance(text, dict) or isinstance(array, dict):
        log_dict(text, array, indent=indent)
        return
    if isinstance(array, (list, tuple)):
        d = dict([(str(i), v) for i, v in enumerate(array)])
        log_dict(text, d, indent=indent)
        return
    if array is not None:
        text = indent + text.ljust(log_options['max_name_length'])
        # if scalar
        if isinstance(array, scalar_types):
            text += f"{array}"
        else:
            try:
                s_mean = '{:10.5f}'.format(array.mean()) if array.size else ""
            except:
                try:
                    s_mean = '{:10.5f}'.format(
                        np.array(array).mean()) if array.size else ""
                except:
                    try:
                        s_mean = '{:10.5f}'.format(array.float().mean()) if array.size else ""
                    except:
                        s_mean = ""
            s_min = calc_and_format_minmax(array, 'min')
            s_max = calc_and_format_minmax(array, 'max')

            dtype = str(array.dtype)
            s_shape = '{:' + str(log_options['max_shape_length']) + '}'
            s_shape = s_shape.format(str(tuple(array.shape)))
            text += ("shape: {}  dtype: {:13}  min: {},  max: {},  mean: {}".format(
                s_shape,
                dtype,
                s_min,
                s_max,
                s_mean))
    print(text)


def log_as_dict(text, array=None, indent=''):
    r = {}
    r['title'] = indent + text
    if array is not None:
        if isinstance(array, scalar_types):
            r['value'] = array
            r['dtype'] = str(type(array))
        else:
            r['dtype'] = str(array.dtype)
            try:
                s_mean = '{:10.5f}'.format(array.mean()) if array.size else ""
            except:
                try:
                    s_mean = '{:10.5f}'.format(
                        np.array(array).mean()) if array.size else ""
                except:
                    s_mean = ""
            r['mean'] = s_mean
            r['min'] = calc_and_format_minmax(array, 'min')
            r['max'] = calc_and_format_minmax(array, 'max')
            # s_shape = '{:' + str(log_options['max_shape_length']) + '}'
            r['shape'] = str(tuple(array.shape))
    return r


def log_dict(text, d=None, indent=''):
    if d is None:
        d = text
    else:
        print(f'{text}:')
    # indent = indent + ' ' * 3
    for key in d:
        try:
            value = d[key]
            s_key = '{:<' + str(log_options['max_name_length']) + '}'
            s_key = s_key.format(key)

            if isinstance(value, str):
                print(f"{indent}{s_key}'{value}'")
            elif isinstance(value, list):
                # print(f"{indent}{s_key}'{value}'")
                d_list = dict([(str(i), v) for i, v in enumerate(value)])
                log(key, d_list, indent=indent + ' ' * 3)
            else:
                log(key, value, indent=indent)
        except Exception as e:
            warnings.warn(f"Can't log key='{key}': {e}")
            # warn_always(f"Can't log key='{key}': {e}")


def calc_and_format_minmax(array, fn='min'):
    s = ''
    if getattr(array, 'size', None):
        if fn == 'min':
            v = array.min()
        else:
            v = array.max()
    else:
        return 'xxx'

    if is_array_integer(array):
        s = '{:10}'.format(v)
    else:
        s = '{:10.5f}'.format(v)
    return s


def is_array_integer(a):
    s_integers = ['torch.int16', 'torch.int32', 'torch.int64',
                  'int16', 'int32', 'int64',
                  ]
    return str(a.dtype) in s_integers




from setuptools import setup, find_packages

__version__ = ""
exec(open('goodok_mlu/version.py').read())


setup(
    name='goodok_mlu',
    version=__version__,
    description='Common ML utils',
    author='Alexey U. Gudchenko',
    author_email='proga@goodok.ru',
    url='https://github.com/goodok/ml_utils',
    packages=find_packages(),
    python_requires='>=3.6',
)
